/***************************************************************************
                          params.cpp  -  description
                             -------------------
    begin                : Sun May 8 2005
    copyright            : (C) 2005 by Mirela Andronescu
    email                : andrones@cs.ubc.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "params.h"
#include "common.h"
#include "externs.h"
#include "constants.h"
#include "string.h"
#include "s_specific_functions.h"
#include "s_hairpin_loop.h"
#include "s_stacked_pair.h"
#include "s_internal_loop.h"
#include "s_min_folding.h"
#include "s_sub_folding.h"
#include "simfold.h"

// I played with the following functions to study a bit the parameters

void print_stacking_energies()
// prints the stacking energies
{
  int i,j,k,l;
  for (i=0; i < NUCL; i++)
      for (j=0; j < NUCL; j++)
          for (k=0; k < NUCL; k++)
          {
              for (l=0; l < NUCL; l++)
                  printf ("stack(%c,%c,%c,%c) = %d\n", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), stack[i][j][k][l]);
              printf ("----\n");
          }
}



void print_tstacki_dangling_energies()
// prints the tstacki and dangling energies
{
    int met = 0;
    int i,j,k,l;
    for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
        {
            met = 0;            
            for (i=0; i < NUCL; i++)            
                for (j=0; j < NUCL; j++)
                {
                    if (tstacki[i][j][k][l] < INF)
                    {
                        met = 1;
                        printf ("tstacki(%c,%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), tstacki[i][j][k][l]);
                        printf ("d3(%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), dangle_top[i][j][k]);
                        printf ("d5(%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(l), dangle_bot[i][j][l]);
                        printf ("Diff = %5d\n", tstacki[i][j][k][l] - dangle_top[i][j][k] - dangle_bot[i][j][l]);
                    }    
                }    
            if (met) printf ("---\n");
        }            
}


void print_stack_dangling_energies()
// prints the stack and dangling energies
{
    int met = 0;
    int i,j,k,l;
    for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
        {
            met = 0;            
            for (i=0; i < NUCL; i++)            
                for (j=0; j < NUCL; j++)
                {
                    if (stack[i][j][k][l] < INF)
                    {
                        met = 1;
                        printf ("stack(%c,%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), stack[i][j][k][l]);
                        printf ("d3(%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), dangle_top[i][j][k]);
                        printf ("d5(%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(l), dangle_bot[i][j][l]);
                        printf ("Diff = %5d\n", stack[i][j][k][l] - dangle_top[i][j][k] - dangle_bot[i][j][l]);
                    }    
                }    
            if (met) printf ("---\n");
        }            
}


void print_stack_equation_dangling()
// prints the stack and dangling energies
{
    int met = 0;
    int i,j,k,l;
    float x,y,z,t;
    x = 0.0592;
    y = -7.0973;
    z = -0.4545;
    t = 7.7252;
    for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
        {
            met = 0;            
            for (i=0; i < NUCL; i++)            
                for (j=0; j < NUCL; j++)
                {
                    if (stack[i][j][k][l] < INF)
                    {
                        met = 1;
                        printf ("stack(%c,%c,%c,%c) = %5d\t", int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), stack[i][j][k][l]);
                        printf (" eq = %5.2lf\n", x*dangle_top[i][j][k] + y*dangle_bot[i][j][l] + z*dangle_top[l][k][j] + t*dangle_bot[l][k][i]);
                    }    
                }    
            if (met) printf ("---\n");
        }            
}

void print_int22_tstacki()
// prints the int22 and tstacki energies
{
    int met = 0;
    int i,j,k,l,m,n,o,p;
    for (i=0; i < NUCL; i++)            
        for (j=0; j < NUCL; j++)    
            for (k=0; k < NUCL; k++)
                for (l=0; l < NUCL; l++)
                    for (m=0; m < NUCL; m++)            
                        for (n=0; n < NUCL; n++)    
                            for (o=0; o < NUCL; o++)
                                for (p=0; p < NUCL; p++)
                                {
                                    if (int22[i][j][k][l][m][n][o][p] < INF)
                                    {
                                        printf ("int22(%c,%c,%c,%c,%c,%c,%c,%c) = %5d -- %5d = \t", int_to_nuc(i), int_to_nuc(j), 
                                            int_to_nuc(k), int_to_nuc(l), int_to_nuc(m), int_to_nuc(n), int_to_nuc(o), int_to_nuc(p),
                                            int22[i][j][k][l][m][n][o][p], tstacki[i][j][k][l]+tstacki[n][m][p][o]);
                                        printf ("tstacki(%c,%c,%c,%c) + tstacki(%c,%c,%c,%c) = %5d + %5d\n",    
                                            int_to_nuc(i), int_to_nuc(j), int_to_nuc(k), int_to_nuc(l), 
                                            int_to_nuc(n), int_to_nuc(m), int_to_nuc(p), int_to_nuc(o), tstacki[i][j][k][l], tstacki[n][m][p][o]);
                                    }
                                }
}


/////////////////////////////////////////////////////////////
// next functions are for playing with parameter learning. //
/////////////////////////////////////////////////////////////


int structure_type_index (char type[])
  // Mirela: Nov 23, 2003
  // Given the type as a string, return the index in string_params
{
  int i, found;
  found = 0;
  for (i=0; i < num_params; i++)
    {
      if (strcmp (type, string_params[i]) == 0)
        // Note: this is very slow
        {
          found = 1;
          break;
        }
    }
  if (!found)
    {
      printf ("!!!TYPE NOT found: %s!!!\n", type);
    }
  return i;
}



void count_AU_penalty (int base_i, int base_j, int *counter)
// PRE:  base_i and base_j make a pair
// POST: increment counter at the appropriate position
// Mirela: Nov 23, 2003
{
  int index;
    if ((base_i != C || base_j != G) &&
        (base_i != G || base_j != C))      
      {
        index = structure_type_index ("misc.terminal_AU_penalty");
        counter[index]++;
      }
}


void count_dangling_energy (int *sequence, int link, int i1, int i2, int i3, int i4, int *counter)
//      (   )...(   )
//      i1..i1..i3..i4
// PRE:  (i1, i2) and (i3, i4) are pairs, i2 and i3 are neighbours, i2 < i3
// POST: return dangling energy between i2 and i3
// Mirela: Nov 23, 2003
{
    int energy;
    int d_top, d_bot;
    char type[100];
    int index_top, index_bot;

    d_top = 0;
    d_bot = 0;

    if (i2 != link)
      {
        d_top = MIN (0, IGINF(dangle_top[sequence[i2]]
                          [sequence[i1]]
                          [sequence[i2+1]]));
        sprintf (type, "dangle_top[%d][%d][%d]",sequence[i2], sequence[i1], sequence[i2+1]);
        index_top = structure_type_index (type);
      }
    if (i3-1 != link)
      {
        d_bot = MIN (0, IGINF(dangle_bot[sequence[i4]]
                          [sequence[i3]]
                          [sequence[i3-1]]));
        sprintf (type, "dangle_bot[%d][%d][%d]",sequence[i4], sequence[i3], sequence[i3-1]);
        index_bot = structure_type_index (type);
      }
    if (i2+1 == i3-1 && i2 == link)
      {
        energy = d_bot;
        counter[index_bot]++;
      }
    else if (i2+1 == i3-1 && i3-1 == link)
      {
        energy = d_top;
        counter[index_top]++;
      }
    else if (i2+1 == i3-1)     // see which is smaller
    {
        energy = d_top < d_bot ? d_top : d_bot;
        // NOTE: the comparison of d_top with d_bot is not right!
        if (d_top < d_bot) counter[index_top]++;
        else counter[index_bot]++;
    }
    else if (i2+1 < i3-1)
    {
        energy = d_top + d_bot;
        counter[index_top]++;
        counter[index_bot]++;
    }

    else // if there is no free base between the two branches, return 0
        energy = 0;
    //    return energy;
}

void count_dangling_energy_left (int *sequence, int link, int i1, int i2, int i3, int i4, int *counter)
//      (....(    )   )
//      i1   i3  i4  i2
// PRE:  (i1, i2) and (i3, i4) are pairs, i1 and i3 are neighbours, i3 < i2
// POST: return dangling energy between i1 and i3
// Mirela: Nov 23, 2003
{
    int energy;
    int d_top, d_bot;
    char type[100];
    int index_top, index_bot;
    d_top = 0;
    d_bot = 0;

    // this will be used in multi-loops.
    // add the dangle_top, even if it is positive
    if (i1 != link)
      {
        d_top = MIN (0, IGINF(dangle_top[sequence[i1]]
                          [sequence[i2]]
                          [sequence[i1+1]]));
        sprintf (type, "dangle_top[%d][%d][%d]",sequence[i1], sequence[i2], sequence[i1+1]);
        index_top = structure_type_index (type);
      }
    // in the other parts of the multi-loop, the dangles are added only if they are negative
    if (i3-1 != link)
      {
        d_bot = MIN (0, IGINF(dangle_bot[sequence[i4]]
                          [sequence[i3]]
                          [sequence[i3-1]]));
        sprintf (type, "dangle_bot[%d][%d][%d]",sequence[i4], sequence[i3], sequence[i3-1]);
        index_bot = structure_type_index (type);
      }
    if (i1+1 == i3-1 && i1 == link)
      {
        energy = d_bot;
        counter[index_bot]++;
      }
    else if (i1+1 == i3-1 && i3-1 == link)
      {
        energy = d_top;
        counter[index_top]++;
      }
    else if (i1+1 == i3-1)     // see which is smaller
    {
        energy = d_top < d_bot ? d_top : d_bot;
        // NOTE: the comparison of d_top with d_bot is not right!
        if (d_top < d_bot) counter[index_top]++;
        else counter[index_bot]++;
    }
    else if (i1+1 < i3-1)
    {
        energy = d_top + d_bot;
        counter[index_top]++;
        counter[index_bot]++;
    }
    else // if there is no free base between the two branches, return 0
        energy = 0;
    //    return energy;
}


void count_dangling_energy_right (int *sequence, int link, int i1, int i2, int i3, int i4, int *counter)
//      (    (    )...)
//      i1   i3  i4  i2
// PRE:  (i1, i2) and (i3, i4) are pairs, i1 and i3 are neighbours, i3 < i2
// POST: return dangling energy between i4 and i2
// Mirela: Nov 23, 2003
{
    int energy;
    int d_top, d_bot;
    char type[100];
    int index_top, index_bot;

    d_top = 0;
    d_bot = 0;

    if (i4 != link)
      {
        d_top = MIN (0, IGINF(dangle_top[sequence[i4]]
                          [sequence[i3]]
                          [sequence[i4+1]]));
        sprintf (type, "dangle_top[%d][%d][%d]",sequence[i4], sequence[i3], sequence[i4+1]);
        index_top = structure_type_index (type);
      }
    if (i2-1 != link)
      {
        d_bot = MIN (0, IGINF(dangle_bot[sequence[i1]]
                          [sequence[i2]]
                          [sequence[i2-1]]));
        sprintf (type, "dangle_bot[%d][%d][%d]",sequence[i1], sequence[i2], sequence[i2-1]);
        index_bot = structure_type_index (type);
      }
    if (i4+1 == i2-1 && i4 == link)
      {
        energy = d_bot;
        counter[index_bot]++;
      }
    else if (i4+1 == i2-1 && i2-1 == link)
      {
        energy = d_top;
        counter[index_top]++;
      }
    else if (i4+1 == i2-1)     // see which is smaller
    {
        energy = d_top < d_bot ? d_top : d_bot;
        // NOTE: the comparison of d_top with d_bot is not right!
        if (d_top < d_bot) counter[index_top]++;
        else counter[index_bot]++;
    }
    else if (i4+1 < i2-1)
    {
        energy = d_top + d_bot;
        counter[index_top]++;
        counter[index_bot]++;
    }
    else // if there is no free base between the two branches, return 0
        energy = 0;
    //    return energy;
}



void count_penalty_by_size (int size, char type, int *counter)
// PRE:  size is the size of the loop
//       type is HAIRP or INTER or BULGE
// POST: return the penalty by size of the loop
// Mirela: Nov 23, 2003
{
  char t[100];
  int index;

    int penalty30, penalty;


    // the penalties for size <= MAXLOOP should be read from the file "loop"
    if (size <= MAXLOOP)
    {
        if (type == 'H')
          {
            //return hairpin_penalty_by_size[size];
            sprintf (t, "hairpin_penalty_by_size[%d]", size);
            index = structure_type_index(t);
            counter[index]++;
            return;
          }
        if (type == 'I')
          {
            //return internal_penalty_by_size[size];
            sprintf (t, "internal_penalty_by_size[%d]", size);
            index = structure_type_index(t);
            counter[index]++;
            return;
          }
        //return bulge_penalty_by_size[size];
        sprintf (t, "bulge_penalty_by_size[%d]", size);
        index = structure_type_index(t);
        counter[index]++;
        return;

    }

    // Mirela: Nov 23, 2003
    // We keep the formula 
    // Extrapolation for large loops based on polymer theory 
    // internal, bulge or hairpin loops > 30: dS(T)=dS(30)+param*ln(n/30) 
    // and imcrement for x_penalty_by_size[MAXLOOP] and for misc.param_greater30

    // size > MAXLOOP
    if (type == 'H')
      {
        penalty30 = hairpin_penalty_by_size[MAXLOOP];
        sprintf (t, "hairpin_penalty_by_size[%d]", MAXLOOP);
        index = structure_type_index (t);
        counter[index]++;        
      }
    else if (type == 'I')
      {
        penalty30 = internal_penalty_by_size[MAXLOOP];
        sprintf (t, "internal_penalty_by_size[%d]", MAXLOOP);
        index = structure_type_index (t);
        counter[index]++;
      }
    else
      {
        penalty30 = bulge_penalty_by_size[MAXLOOP];
        sprintf (t, "bulge_penalty_by_size[%d]", MAXLOOP);
        index = structure_type_index (t);
        counter[index]++;
      }

    // by Cho MinSik
    static  double LogValue[INF];
    if(LogValue[size]==0)
        LogValue[size]  =   log(((double)size)/30);

    penalty = (int) (penalty30 + misc.param_greater30 * LogValue[size]);
    index = structure_type_index ("misc.param_greater30");
    // TODO: to store double instead of int
    counter[index]+= (int) LogValue[size];

    //    return penalty;
    return;
}

void count_asymmetry_penalty (int size1, int size2, int *counter)
// PRE:  size1 and size2 are the sizes of the two free base groups
// POST: Calculate the asymmetry penalty for internal loops
//       Note that if size1 == size2, pen is 0
// Mirela: Nov 23, 2003
{
    int ass, pen, index;    
    ass = misc.asymmetry_penalty_array [MIN (2, MIN (size1, size2))-1];
    pen = MIN (misc.asymmetry_penalty_max_correction, abs (size1-size2) * ass);
    // pen can only be 0.5, 1, 1.5, 2, 2.5 or 3 with the current model;
    switch (size1-size2)
      {
        case 1: index = structure_type_index("misc.asymmetry_penalty[1]"); counter[index]++; break;
        case 2: index = structure_type_index("misc.asymmetry_penalty[2]"); counter[index]++; break;
        case 3: index = structure_type_index("misc.asymmetry_penalty[3]"); counter[index]++; break;
        case 4: index = structure_type_index("misc.asymmetry_penalty[4]"); counter[index]++; break;
        case 5: index = structure_type_index("misc.asymmetry_penalty[5]"); counter[index]++; break;
        default:index = structure_type_index("misc.asymmetry_penalty[6]"); counter[index]++; break;          
      }
    
    //    return pen;
}


void count_types (int link, int *sequence, char *csequence, str_features *f, int *counter)
  // Mirela: Nov 23, 2003
  // instead of summing up the energies of each type, increment the corresponding counter
{
    int energy, i, en, AUpen;
    int dang;
    int misc_energy;
    int h,l;    
    static int cannot_add_dangling[MAXSLEN];
    char type[100];
    int index;
    
    int nb_nucleotides = strlen(csequence);
    for (i=0; i < nb_nucleotides; i++) cannot_add_dangling[i] = 0;

    energy = 0;
    AUpen = 0;
    
    for (i=0; i < nb_nucleotides; i++)
    {
    
        // add some AU_penalties
        if ((i==0 || (i-1 == link && !cannot_add_dangling[i-1] && f[i-1].pair == -1))
             && f[i].pair > i)
        {
            AUpen = AU_penalty (sequence[i], sequence[f[i].pair]);
            if (debug)
                printf ("%d - AUpen1 \t- add energy %6d\n", i, AUpen);            
            energy += AUpen;            
            count_AU_penalty (sequence[i], sequence[f[i].pair], counter);
        }            
        else if ( i > 0 && f[i].pair > i && f[i-1].pair < i-1 &&
             f[i-1].pair != -1 && !cannot_add_dangling[i])
            //  )(  
        {            
            AUpen = AU_penalty (sequence[i], sequence[f[i].pair]);
            if (debug)
                printf ("%d - AUpen2 \t- add energy %6d\n", i, AUpen);
            energy += AUpen; 
            count_AU_penalty (sequence[i], sequence[f[i].pair], counter);
        }            
    
        // add dangling energies and AU_penalties
        if (f[i].pair == -1 && !cannot_add_dangling[i])
        {
            if ((i == 0 || i-1 == link || (i > 0 && f[i-1].pair == -1 && i != link)) &&
                 i < nb_nucleotides-1 && f[i+1].pair > i+1)
                // .( or ..(
            {
                dang = MIN (0, IGINF(dangle_bot [sequence[f[i+1].pair]] [sequence[i+1]] [sequence[i]]));
                AUpen = AU_penalty (sequence[i+1], sequence[f[i+1].pair]);
                if (debug)
                {
                    printf ("%d - dangle1 \t- add energy %6d\n", i, dang);
                    printf ("%d - AUpen3 \t- add energy %6d\n", i, AUpen);
                }                    
                energy += dang + AUpen;
                sprintf (type, "dangle_bot[%d][%d][%d]",sequence[f[i+1].pair], sequence[i+1], sequence[i]);
                index = structure_type_index (type);
                counter[index]++;
                count_AU_penalty (sequence[i+1], sequence[f[i+1].pair], counter);
            }
            else if ((i == nb_nucleotides-1 || i == link ||
                     (i < nb_nucleotides-1 && f[i+1].pair == -1 && i-1 != link)) &&
                     i > 0 && f[i-1].pair > -1 && f[i-1].pair < i-1)
                // ). or )..
            {
                dang = MIN (0, IGINF(dangle_top [sequence[i-1]] [sequence[f[i-1].pair]] [sequence[i]]));
                if (debug)                
                    printf ("%d - dangle2 \t- add energy %6d\n", i, dang);
                energy += dang;
                sprintf (type, "dangle_top[%d][%d][%d]",sequence[i-1], sequence[f[i-1].pair], sequence[i]);
                index = structure_type_index (type);
                counter[index]++;
            }
            else if (i < nb_nucleotides-1 && f[i+1].pair > i+1 && f[i-1].pair < i-1 && f[i-1].pair != -1)
               // ).( 
            {
                dang = MIN (0, IGINF(s_dangling_energy (sequence, f[i-1].pair, i-1, i+1, f[i+1].pair)));
                AUpen = AU_penalty (sequence[i+1], sequence[f[i+1].pair]);
                if (debug)
                {              
                    printf ("%d - dangle1 \t- add energy %6d\n", i, dang);
                    printf ("%d - AUpen4 \t- add energy %6d\n", i, AUpen);
                }    
                energy += dang + AUpen;
                count_dangling_energy (sequence, link, f[i-1].pair, i-1, i+1, f[i+1].pair, counter);
                count_AU_penalty (sequence[i+1], sequence[f[i+1].pair], counter);
            }
            else
            {
                continue;
            }
        }
        
        if (f[i].pair < i)
        {
            continue;
        }       

        if (f[i].type == STACK)
        {
            en = s_stacked_pair::get_energy (i, f[i].pair, sequence);
            if (debug)            
                printf ("%d stack \t- add energy %6d\n", i, en);        
            energy += en;
            s_stacked_pair::count_get_energy (i, f[i].pair, sequence, counter);
        }
        else if (f[i].type == HAIRP)
        {
            en = s_hairpin_loop::get_energy (i, f[i].pair, sequence, csequence);
            if (debug)
                printf ("%d hairpin \t- add energy %6d\n", i, en);
            energy += en;
            s_hairpin_loop::count_get_energy (i, f[i].pair, sequence, csequence, counter);
        }
        else if (f[i].type == INTER)
        {
            int ip, jp;
            ip = f[i].bri[0];
            jp = f[f[i].bri[0]].pair;
            cannot_add_dangling[ip-1] = 1;
            cannot_add_dangling[jp+1] = 1;
            en = s_internal_loop::get_energy (i, f[i].pair, ip, jp, sequence);
            if (debug)
                printf ("%d internal \t- add energy %6d\n", i, en);        
            energy += en;
            s_internal_loop::count_get_energy (i, f[i].pair, ip, jp, sequence, counter);
        }
        else  // (f[i].type == MULTI)
        {
            dang = 0;
            misc_energy = 0;
            AUpen = 0;
            int special;
            special = 0;
            // add the energies/enthalpies for free bases
            // first find out if it is a regular multi-loop or a special multi-loop
            l = i;

            while (l < f[i].bri[0] && !special)
                if (l++ == link) special = 1;
            h = 0;                
            while (h < f[i].num_branches-1 && !special)
            {
                l = f[f[i].bri[h]].pair;
                while (l < f[i].bri[h+1] && !special)
                    if (l++ == link) special = 1;
                h++;    
            }
            l = f[f[i].bri[f[i].num_branches-1]].pair;
            while (l < f[i].pair && !special)
                if (l++ == link) special = 1;
            // now we now if it a special multi-loop or not                                  

            if (!special)
            {
//                printf ("Regular ML\n");
              index = structure_type_index ("misc.multi_free_base_penalty");
                for (l=i+1; l < f[i].bri[0]; l++)
                  {
                    misc_energy += misc.multi_free_base_penalty;
                    counter[index]++;
                  }
                for (h=0; h < f[i].num_branches-1; h++)
                {
                    for (l = f[f[i].bri[h]].pair + 1; l < f[i].bri[h+1]; l++)
                      {
                        misc_energy += misc.multi_free_base_penalty;
                        counter[index]++;
                      }
                }
                for (l = f[f[i].bri[f[i].num_branches-1]].pair + 1; l < f[i].pair; l++)
                  {
                    misc_energy += misc.multi_free_base_penalty;
                    counter[index]++;
                  }
                misc_energy += misc.multi_offset;
                misc_energy += misc.multi_helix_penalty * (f[i].num_branches + 1);
                index = structure_type_index ("misc.multi_offset");
                counter[index]++;
                index = structure_type_index ("misc.multi_helix_penalty");
                counter[index]+= f[i].num_branches + 1;
            }
            else
            {
//                printf ("Special ML\n");            
                misc_energy = misc.intermolecular_initiation;
                index = structure_type_index ("misc.intermolecular_initiation");
                counter[index]++;
            }                
            // add AU_penalties for multi-loop
            AUpen += AU_penalty (sequence[i], sequence[f[i].pair]);
            count_AU_penalty (sequence[i], sequence[f[i].pair], counter);
            for (h=0; h < f[i].num_branches; h++)
              {
                AUpen += AU_penalty (sequence[f[i].bri[h]],sequence[f[f[i].bri[h]].pair]);
                count_AU_penalty (sequence[f[i].bri[h]],sequence[f[f[i].bri[h]].pair], counter);
              }
        
            // add dangling energies for multi-loop
            dang += s_dangling_energy_left (sequence, i, f[i].pair, f[i].bri[0], f[f[i].bri[0]].pair);
            count_dangling_energy_left (sequence, link, i, f[i].pair, f[i].bri[0], f[f[i].bri[0]].pair, counter);
            for (l=0; l < f[i].num_branches - 1; l++)
              {
                dang += s_dangling_energy (sequence, f[i].bri[l], f[f[i].bri[l]].pair, f[i].bri[l+1], f[f[i].bri[l+1]].pair);
                count_dangling_energy (sequence, link, f[i].bri[l], f[f[i].bri[l]].pair, f[i].bri[l+1], f[f[i].bri[l+1]].pair, counter);
              }
            dang += s_dangling_energy_right (sequence, i, f[i].pair, f[i].bri[f[i].num_branches-1], f[f[i].bri[f[i].num_branches-1]].pair);
            count_dangling_energy_right (sequence, link, i, f[i].pair, f[i].bri[f[i].num_branches-1], f[f[i].bri[f[i].num_branches-1]].pair, counter);

            // add "no-dangling" restriction                                    
            for (l=0; l < f[i].num_branches; l++)
            {
                cannot_add_dangling [f[i].bri[l] -1] = 1;
                cannot_add_dangling [f[f[i].bri[l]].pair + 1] = 1;
            }
            if (debug)
            {
                printf ("%d - multi m\t- add energy %6d\n", i, misc_energy);            
                printf ("%d - multi d\t- add energy %6d\n", i, dang);
                printf ("%d - multi AU\t- add energy %6d\n", i, AUpen);
            }                
            energy += misc_energy + dang + AUpen;                           
        }
    }
    //    return energy;
}


void count_each_structure_type (char *sequence, char *structure, int *counter)
// Mirela: Nov 23, 2003
// GIven the sequence and the structure, returns the vector counter, with the #of elementary structures
{
    int i;
    str_features *f;
    int *int_sequence;

    //first reset counter
    for (i=0; i < num_params; i++)
    {
        counter[i] = 0;
    }
    
    int nb_nucleotides = strlen(sequence);
    
    if ((f = new str_features[nb_nucleotides]) == NULL) giveup ("Cannot allocate memory", "str_features");
    nb_nucleotides = strlen(sequence);    
    // detect the structure features
    detect_structure_features (structure, f);
    
    int_sequence = new int[nb_nucleotides];
    if (int_sequence == NULL) giveup ("Cannot allocate memory", "energy");
    for (i=0; i < nb_nucleotides; i++) int_sequence[i] = nuc_to_int(sequence[i]);
    count_types (-1, int_sequence, sequence, f, counter);
    delete [] int_sequence;
    delete [] f;    
}



int create_string_params ()
  // Mirela: Nov 23, 2003
  // writes each parameter type, excluding duplicates, in a long vector, containing the names of the parameters
{
  int index;
  int i, j, k, l, m, n, o, p;
  index = 0;
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (stack[i][j][k][l] < INF)
              {
                // exclude duplicates
                // stack[i][j][k][l] is the same as stack[l][k][j][i]
                if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                  {
                    sprintf (string_params[index++], "stack[%d][%d][%d][%d]", i, j, k, l);
                  }
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstackh[i][j][k][l] < INF)
              {
                // no duplicates here
                sprintf (string_params[index++], "tstackh[%d][%d][%d][%d]", i, j, k, l);
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstacki[i][j][k][l] < INF)
              {
                // no duplicates here
                sprintf (string_params[index++], "tstacki[%d][%d][%d][%d]", i, j, k, l);
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              {
                if (int11[i][j][k][l][m][n] < INF)
                  {
                    // exclude duplicates
                    // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                    if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                      {
                        sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                      }
                  }
              }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              for(o=0; o < NUCL; o++)
                {
                  if (int21[i][j][k][l][m][n][o] < INF)
                    {
                      // no duplicates here
                      sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                    }
                }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              for(o=0; o < NUCL; o++)
                for (p=0; p < NUCL; p++)
                  {
                    if (int22[i][j][k][l][m][n][o][p] < INF)
                      {
                        // exclude duplicates
                        // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                        if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                            n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                          {
                            sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                          }
                      }
                  }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_top[i][j][k] < INF)
            {
              // no duplicates here
              sprintf (string_params[index++], "dangle_top[%d][%d][%d]", i, j, k);
            }
        }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_bot[i][j][k] < INF)
            {
              // no duplicates here
              sprintf (string_params[index++], "dangle_bot[%d][%d][%d]", i, j, k);
            }
        }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (internal_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index++], "internal_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (bulge_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index++], "bulge_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (hairpin_penalty_by_size[i] < INF)
        {
          // no duplicates here
          sprintf (string_params[index++], "hairpin_penalty_by_size[%d]", i);
        }
    }

 sprintf (string_params[index++], "misc.param_greater30");
 sprintf (string_params[index++], "misc.terminal_AU_penalty");
 sprintf (string_params[index++], "misc.hairpin_GGG");
 sprintf (string_params[index++], "misc.hairpin_c1");
 sprintf (string_params[index++], "misc.hairpin_c2");
 sprintf (string_params[index++], "misc.hairpin_c3");
 // sprintf (string_params[index++], "misc.asymmetry_penalty_max_correction");
 // sprintf (string_params[index++], "misc.asymmetry_penalty_array[0]");
 // sprintf (string_params[index++], "misc.asymmetry_penalty_array[1]");
 // sprintf (string_params[index++], "misc.asymmetry_penalty_array[2]");
 // sprintf (string_params[index++], "misc.asymmetry_penalty_array[3]");
 // Instead of these, I will just store the asymmetry for 0.5, 1, 1.5, 2, 2.5 and 3.
 sprintf (string_params[index++], "misc.asymmetry_penalty[1]");
 sprintf (string_params[index++], "misc.asymmetry_penalty[2]");
 sprintf (string_params[index++], "misc.asymmetry_penalty[3]");
 sprintf (string_params[index++], "misc.asymmetry_penalty[4]");
 sprintf (string_params[index++], "misc.asymmetry_penalty[5]");
 sprintf (string_params[index++], "misc.asymmetry_penalty[6]");

 sprintf (string_params[index++], "misc.gail_rule");
 sprintf (string_params[index++], "misc.multi_offset");
 sprintf (string_params[index++], "misc.multi_helix_penalty");
 sprintf (string_params[index++], "misc.multi_free_base_penalty");
 sprintf (string_params[index++], "misc.intermolecular_initiation");

 for(i=0; i < nb_triloops; i++)
   {
     sprintf (string_params[index++], "triloop[%d].energy", i);
   }

 for(i=0; i < nb_tloops; i++)
   {
     sprintf (string_params[index++], "tloop[%d].energy", i);
   }
 return index;
}




void fill_data_structures_with_new_parameters (char *filename)
  // Mirela: Dec 16, 2003
  // reads parameters from a file, and writes them in the internal data structures
  // PRE: first read the actual standard parameters, to be able to figure out which of them are
  // < INF, and maybe to also keep some old values.
{
  int index;
  int i, j, k, l, m, n, o, p;
  FILE *file;
  char buffer[100];
  double param;

  if ((file = fopen (filename, "r")) == NULL)
    {
      giveup ("Cannot open file", filename);
    }

  index = 0;
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (stack[i][j][k][l] < INF)
              {
                // exclude duplicates
                // stack[i][j][k][l] is the same as stack[l][k][j][i]
                if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                  {
                    fgets (buffer, sizeof(buffer), file);
                    sscanf (buffer, "%lf\n", &param);
                    if (param != 0)
                      {
                        param *= 100;
                        stack[i][j][k][l] = (int)param;
                        // add the duplicate too
                        stack[l][k][j][i] = (int)param;
                      }
                    //sprintf (string_params[index++], "stack[%d][%d][%d][%d]", i, j, k, l);
                  }
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstackh[i][j][k][l] < INF)
              {
                // no duplicates here
                fgets (buffer, sizeof(buffer), file);
                sscanf (buffer, "%lf", &param);
                if (param != 0)
                  {
                    param *= 100;
                    tstackh[i][j][k][l] = (int)param;
                  }
                //sprintf (string_params[index++], "tstackh[%d][%d][%d][%d]", i, j, k, l);
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstacki[i][j][k][l] < INF)
              {
                // no duplicates here
                fgets (buffer, sizeof(buffer), file);
                sscanf (buffer, "%lf", &param);
                if (param != 0)
                  {
                    param *= 100;
                    tstacki[i][j][k][l] = (int)param;
                  }
                //sprintf (string_params[index++], "tstacki[%d][%d][%d][%d]", i, j, k, l);
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              {
                if (int11[i][j][k][l][m][n] < INF)
                  {
                    // exclude duplicates
                    // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                    if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                      {
                        fgets (buffer, sizeof(buffer), file);
                        sscanf (buffer, "%lf", &param);
                        if (param != 0)
                          {
                            param *= 100;
                            int11[i][j][k][l][m][n] = (int)param;
                            // add the duplicate too
                            int11[n][m][l][k][j][i] = (int)param;
                          }
                        //sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                      }
                  }
              }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              for(o=0; o < NUCL; o++)
                {
                  if (int21[i][j][k][l][m][n][o] < INF)
                    {
                      // no duplicates here
                      fgets (buffer, sizeof(buffer), file);
                      sscanf (buffer, "%lf", &param);
                      if (param != 0)
                        {
                          param *= 100;
                          int21[i][j][k][l][m][n][o] = (int)param;
                        }
                      //sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                    }
                }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              for(o=0; o < NUCL; o++)
                for (p=0; p < NUCL; p++)
                  {
                    if (int22[i][j][k][l][m][n][o][p] < INF)
                      {
                        // exclude duplicates
                        // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                        if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                            n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                          {
                            fgets (buffer, sizeof(buffer), file);
                            sscanf (buffer, "%lf", &param);
                            if (param != 0)
                              {
                                param *= 100;
                                int22[i][j][k][l][m][n][o][p] = (int)param;
                                // add the duplicate too
                                int22[n][m][p][0][j][i][l][k] = (int)param;
                              }
                            //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                          }
                      }
                  }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_top[i][j][k] < INF)
            {
              // no duplicates here
              fgets (buffer, sizeof(buffer), file);
              sscanf (buffer, "%lf", &param);
              if (param != 0)
                {
                  param *= 100;
                  dangle_top[i][j][k] = (int)param;
                }
              //sprintf (string_params[index++], "dangle_top[%d][%d][%d]", i, j, k);
            }
        }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_bot[i][j][k] < INF)
            {
              // no duplicates here
              fgets (buffer, sizeof(buffer), file);
              sscanf (buffer, "%lf", &param);
              if (param != 0)
                {
                  param *= 100;
                  dangle_bot[i][j][k] = (int)param;
                }
              //sprintf (string_params[index++], "dangle_bot[%d][%d][%d]", i, j, k);
            }
        }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (internal_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fgets (buffer, sizeof(buffer), file);
          sscanf (buffer, "%lf", &param);
          if (param != 0)
            {
              param *= 100;
              internal_penalty_by_size[i] = (int)param;
            }
          //sprintf (string_params[index++], "internal_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (bulge_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fgets (buffer, sizeof(buffer), file);
          sscanf (buffer, "%lf", &param);
          if (param != 0)
            {
              param *= 100;
              bulge_penalty_by_size[i] = (int)param;
            }
          //sprintf (string_params[index++], "bulge_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (hairpin_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fgets (buffer, sizeof(buffer), file);
          sscanf (buffer, "%lf", &param);
          if (param != 0)
            {
              param *= 100;
              hairpin_penalty_by_size[i] = (int)param;
            }
          //sprintf (string_params[index++], "hairpin_penalty_by_size[%d]", i);
        }
    }
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.param_greater30 = (int)param;
  //sprintf (string_params[index++], "misc.param_greater30");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.terminal_AU_penalty = (int)param;
  //sprintf (string_params[index++], "misc.terminal_AU_penalty");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.hairpin_GGG = (int)param;
  //sprintf (string_params[index++], "misc.hairpin_GGG");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.hairpin_c1 = (int)param;
  //sprintf (string_params[index++], "misc.hairpin_c1");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.hairpin_c2 = (int)param;
  //sprintf (string_params[index++], "misc.hairpin_c2");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.hairpin_c3 = (int)param;
  //sprintf (string_params[index++], "misc.hairpin_c3");

  // sprintf (string_params[index++], "misc.asymmetry_penalty_max_correction");
  // sprintf (string_params[index++], "misc.asymmetry_penalty_array[0]");
  // sprintf (string_params[index++], "misc.asymmetry_penalty_array[1]");
  // sprintf (string_params[index++], "misc.asymmetry_penalty_array[2]");
  // sprintf (string_params[index++], "misc.asymmetry_penalty_array[3]");
  // Instead of these, I will just store the asymmetry for 0.5, 1, 1.5, 2, 2.5 and 3.

  // to come back!!!
  fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
  fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
  fgets (buffer, sizeof(buffer), file); fgets (buffer, sizeof(buffer), file);
  //sprintf (string_params[index++], "misc.asymmetry_penalty[1]");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[2]");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[3]");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[4]");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[5]");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[6]");

  fgets (buffer, sizeof(buffer), file);
  //sprintf (string_params[index++], "misc.gail_rule");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.multi_offset = (int)param;
  //sprintf (string_params[index++], "misc.multi_offset");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.multi_helix_penalty = (int)param;
  //sprintf (string_params[index++], "misc.multi_helix_penalty");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.multi_free_base_penalty = (int)param;
  //sprintf (string_params[index++], "misc.multi_free_base_penalty");
  fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
  if (param != 0)  misc.intermolecular_initiation = (int)param;
  //sprintf (string_params[index++], "misc.intermolecular_initiation");

  for(i=0; i < nb_triloops; i++)
    {
      fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
      if (param != 0)  triloop[i].energy = (int)param;
      //sprintf (string_params[index++], "triloop[%d].energy", i);
    }
  
  for(i=0; i < nb_tloops; i++)
    {
      fgets (buffer, sizeof(buffer), file); sscanf (buffer, "%lf", &param); param *= 100;
      if (param != 0)  tloop[i].energy = (int)param;
      //sprintf (string_params[index++], "tloop[%d].energy", i);
    }
  
  fclose (file);
}



void save_parameters (char *filename)
  // Mirela: Dec 16, 2003
  // save all parameters in the given file
{
  int index;
  int i, j, k, l, m, n, o, p;
  FILE *file;

  if ((file = fopen (filename, "w")) == NULL)
    {
      giveup ("Cannot open file", filename);
    }

  index = 0;
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (stack[i][j][k][l] < INF)
              {
                // exclude duplicates
                // stack[i][j][k][l] is the same as stack[l][k][j][i]
                if (i*1000 + j*100 + k*10 + l <= l*1000 + k*100 + j*10 + i)
                  {
                    fprintf (file, "%.2lf\n", stack[i][j][k][l]/100.0);
                  }
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstackh[i][j][k][l] < INF)
              {
                // no duplicates here
                fprintf (file, "%.2lf\n", tstackh[i][j][k][l]/100.0);
                //sprintf (string_params[index++], "tstackh[%d][%d][%d][%d]", i, j, k, l);
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          {
            if (tstacki[i][j][k][l] < INF)
              {
                // no duplicates here
                fprintf (file, "%.2lf\n", tstacki[i][j][k][l]/100.0);
                //sprintf (string_params[index++], "tstacki[%d][%d][%d][%d]", i, j, k, l);
              }
          }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              {
                if (int11[i][j][k][l][m][n] < INF)
                  {
                    // exclude duplicates
                    // int11[i][j][k][l][m][n] is the same as int11[n][m][l][k][j][i]
                    if (i*100000 + j*10000 + k*1000 + l*100 + m*10 + n <= n*100000 + m*10000 + l*1000+ k*100 + j*10 + i)
                      {
                        fprintf (file, "%.2lf\n", int11[i][j][k][l][m][n]/100.0);
                        //sprintf (string_params[index++], "int11[%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n);
                      }
                  }
              }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              for(o=0; o < NUCL; o++)
                {
                  if (int21[i][j][k][l][m][n][o] < INF)
                    {
                      // no duplicates here
                      fprintf (file, "%.2lf\n", int21[i][j][k][l][m][n][o]/100.0);
                      //sprintf (string_params[index++], "int21[%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o);
                    }
                }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        for (l=0; l < NUCL; l++)
          for (m=0; m < NUCL; m++)
            for (n=0; n < NUCL; n++)
              for(o=0; o < NUCL; o++)
                for (p=0; p < NUCL; p++)
                  {
                    if (int22[i][j][k][l][m][n][o][p] < INF)
                      {
                        // exclude duplicates
                        // int22[i][j][k][l][m][n][o][p] is the same as int22[n][m][p][o][j][i][l][k]
                        if (i*10000000 + j*1000000 + k*100000 + l*10000 + m*1000 + n*100 + o*10 + p <= 
                            n*10000000 + m*1000000 + p*100000 + o*10000 + j*1000 + i*100 + l*10 + k)
                          {
                            fprintf (file, "%.2lf\n", int22[i][j][k][l][m][n][o][p]/100.0);
                            //sprintf (string_params[index++], "int22[%d][%d][%d][%d][%d][%d][%d][%d]", i, j, k, l, m, n, o, p);
                          }
                      }
                  }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_top[i][j][k] < INF)
            {
              // no duplicates here
              fprintf (file, "%.2lf\n", dangle_top[i][j][k]/100.0);
              //sprintf (string_params[index++], "dangle_top[%d][%d][%d]", i, j, k);
            }
        }
  for (i=0; i < NUCL; i++)
    for (j=0; j < NUCL; j++)
      for (k=0; k < NUCL; k++)
        {
          if (dangle_bot[i][j][k] < INF)
            {
              // no duplicates here
              fprintf (file, "%.2lf\n", dangle_bot[i][j][k]/100.0);
              //sprintf (string_params[index++], "dangle_bot[%d][%d][%d]", i, j, k);
            }
        }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (internal_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fprintf (file, "%.2lf\n", internal_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "internal_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (bulge_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fprintf (file, "%.2lf\n", bulge_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "bulge_penalty_by_size[%d]", i);
        }
    }
  for (i=1; i <= MAXLOOP; i++)
    {
      if (hairpin_penalty_by_size[i] < INF)
        {
          // no duplicates here
          fprintf (file, "%.2lf\n", hairpin_penalty_by_size[i]/100.0);
          //sprintf (string_params[index++], "hairpin_penalty_by_size[%d]", i);
        }
    }
  fprintf (file, "%.2lf\n", misc.param_greater30/100.0);
  //sprintf (string_params[index++], "misc.param_greater30");
  fprintf (file, "%.2lf\n", misc.terminal_AU_penalty/100.0);
  //sprintf (string_params[index++], "misc.terminal_AU_penalty");
  fprintf (file, "%.2lf\n", misc.hairpin_GGG/100.0);
  //sprintf (string_params[index++], "misc.hairpin_GGG");
  fprintf (file, "%.2lf\n", misc.hairpin_c1/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c1");
  fprintf (file, "%.2lf\n", misc.hairpin_c2/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c2");
  fprintf (file, "%.2lf\n", misc.hairpin_c3/100.0);
  //sprintf (string_params[index++], "misc.hairpin_c3");
  fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[1]");
  fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[2]");
  fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[3]");
  fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[4]");
  fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[5]");
  fprintf (file, "0\n");
  //sprintf (string_params[index++], "misc.asymmetry_penalty[6]");  
  fprintf (file, "%.2lf\n", misc.gail_rule);
  //sprintf (string_params[index++], "misc.gail_rule");
  fprintf (file, "%.2lf\n", misc.multi_offset/100.0);
  //sprintf (string_params[index++], "misc.multi_offset");
  fprintf (file, "%.2lf\n", misc.multi_helix_penalty/100.0);
  //sprintf (string_params[index++], "misc.multi_helix_penalty");
  fprintf (file, "%.2lf\n", misc.multi_free_base_penalty/100.0);
  //sprintf (string_params[index++], "misc.multi_free_base_penalty");
  fprintf (file, "%.2lf\n", misc.intermolecular_initiation/100.0);
  //sprintf (string_params[index++], "misc.intermolecular_initiation");

 for(i=0; i < nb_triloops; i++)
   {
     fprintf (file, "%.2lf\n", triloop[i].energy/100.0);
     //sprintf (string_params[index++], "triloop[%d].energy", i);
   }

 for(i=0; i < nb_tloops; i++)
   {
     fprintf (file, "%.2lf\n", tloop[i].energy/100.0);
     //sprintf (string_params[index++], "tloop[%d].energy", i);
   }

 fclose (file);
}


double simfold_restricted_logZ (char *sequence, char *real_structure, char *restricted, double &min_energy, double &max_energy, int &actual_num_str)
// if the real_structure is not one of the suboptimal structures, add it 
{
    char structure[MAXSLEN];
    double enthalpy, energy;
    char tmp_structures[MAXSUBSTR][MAXSLEN];
    double tmp_energies[MAXSUBSTR];
    double Z;
    int i;
    double R, temp, beta;
    int real_str_found;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp);
    real_str_found = 0;
    
    
    s_min_folding *min_fold = new s_min_folding (sequence, restricted);
    min_energy = min_fold->s_simfold_restricted ();
    min_fold->return_structure (structure);
    delete min_fold;      
    
    s_sub_folding* sub_fold = new s_sub_folding(sequence, restricted, -(int)(min_energy*100.0));
    sub_fold->set_limit (MAXSUBSTR);
    sub_fold->s_simfold_restricted (enthalpy);
    actual_num_str = sub_fold->return_structures(tmp_structures, tmp_energies);
    delete sub_fold;
    
    max_energy = tmp_energies[actual_num_str-1];
    Z = 0;
    for (i=0; i < actual_num_str; i++)
    {
        // check if the real structure is one of the suboptimal ones
        if (strcmp (tmp_structures[i], real_structure) == 0)
            real_str_found = 1;
        // recompute the free energy, i.e. with the correct dangling ends
        energy = free_energy_simfold_restricted (sequence, tmp_structures[i], restricted);
        Z += exp ((-1) * energy * beta);
    }
    if (!real_str_found)
    {
        // if not found, add it
        energy = free_energy_simfold_restricted (sequence, real_structure, restricted);
        Z += exp ((-1) * energy * beta);        
    }        
    return log(Z);    
}


double simfold_restricted_logZ_gradient (char *sequence, char *real_structure, char *restricted, double *logZ_gradient)
// return 
// if the real_structure is not one of the suboptimal structures, add it 
{
    char structure[MAXSLEN];
    double enthalpy, energy;
    char tmp_structures[MAXSUBSTR][MAXSLEN];
    double tmp_energies[MAXSUBSTR];
    double min_energy, max_energy;
    int i, k;
    double R, temp, beta;
    int real_str_found, actual_num_str;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp);
    real_str_found = 0;
    double numerator [MAXNUMPARAMS];
    int counter [MAXNUMPARAMS];
    double denominator;    
    
    s_min_folding *min_fold = new s_min_folding (sequence, restricted);
    min_energy = min_fold->s_simfold_restricted ();
    min_fold->return_structure (structure);
    delete min_fold;      
    
    s_sub_folding* sub_fold = new s_sub_folding(sequence, restricted, -(int)(min_energy*100.0));
    sub_fold->set_limit (MAXSUBSTR);
    sub_fold->s_simfold_restricted (enthalpy);
    actual_num_str = sub_fold->return_structures(tmp_structures, tmp_energies);
    max_energy = tmp_energies[actual_num_str-1];
    delete sub_fold;
    
    for (i=0; i < num_params; i++)
    {
        numerator[i] = 0;
    }    
        
    // first compute the denominator, which is the same for all parameters
    denominator = 0;
    for (k=0; k < actual_num_str; k++)
    {
        // check if the real structure is one of the suboptimal ones
        if (strcmp (tmp_structures[k], real_structure) == 0)
            real_str_found = 1;
        // recompute the free energy, i.e. with the correct dangling ends
        energy = free_energy_simfold_restricted (sequence, tmp_structures[k], restricted);        
        denominator += exp ((-1) * energy * beta);
        count_each_structure_type (sequence, tmp_structures[k], counter);
        for (i=0; i < num_params; i++)
        {
            numerator[i] += counter[i] * exp ((-1) * energy * beta);
        }
        //printf ("Sub str %d, counter[i] = %d, numerator[i] = %e\n", k, counter[7646], numerator[7646]);
    }
    if (!real_str_found)
    {
        //printf ("Real str not found, add it\n");
        // if not found, add it
        energy = free_energy_simfold_restricted (sequence, real_structure, restricted);
        denominator += exp ((-1) * energy * beta);        
        count_each_structure_type (sequence, real_structure, counter);
        for (i=0; i < num_params; i++)
        {
            numerator[i] += counter[i] * exp ((-1) * energy * beta);
        }    
    }
    // now the denominator and nominator are computed
    
    for (i=0; i < num_params; i++)
    {
        logZ_gradient[i] = numerator[i] / denominator;
    }      
    //printf ("logZ_gradient[i] = %e\n", logZ_gradient[7646]);
}


double compute_probability (double energy, double Z)
{
    double pb;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp);
    pb = exp ((-1) * energy * beta) / Z;
    return pb;
}

int get_info_from_file (FILE *file, char *sequence, char *real_structure, char *restricted)
// return 1 if everything was ok
// return 0 instead of continue;
{
    char buffer [5000];
    int i;
    fgets (buffer, sizeof(buffer), file);
    if (strstr (buffer, "Seq: ") != NULL)
    {
        for (i=5; i < strlen (buffer)-1; i++)
            sequence[i-5] = buffer[i]; 
        sequence[i-5] = '\0';
        // ignore sequences longer than MAXSLEN -1 
        if (strlen (sequence) > MAXSLEN-1) return 0;
    }
    else return 0;
    fgets (buffer, sizeof(buffer), file);
    if (strstr (buffer, "Str: ") != NULL)
    {
        for (i=5; i < strlen (buffer)-1; i++)
            real_structure[i-5] = buffer[i]; 
        real_structure[i-5] = '\0';
    }        
    else
    {
        printf ("Str doesn't follow Seq, Seq is %s\n", sequence);
        exit(1);
    }
    fgets (buffer, sizeof(buffer), file);
    if (strstr (buffer, "Res: ") != NULL)
    {
        for (i=5; i < strlen (buffer)-1; i++)
            restricted[i-5] = buffer[i]; 
        restricted[i-5] = '\0';
    }        
    else
    {
        printf ("Res doesn't follow Str\n");
        exit(1);
    }        
    return 1;
}

double compute_f (char *input_file)
// PRE:  the parameters have been read
// POST: computes the f function, which is sum_{i=1}^N (1/RT G_{i,nat,theta} + logZ)
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double f; 
    double min_energy, max_energy;      
    int i, actual_num_str;
        
    f = 0;
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }
      
    int k;
    k = 0;
    while (!feof (file))
    {
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;
        //printf ("      ....,....1....,....2....,....3....,....4....,....5\n");
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        energy = free_energy_simfold_restricted (sequence, real_structure, restricted);        
        logZ = simfold_restricted_logZ (sequence, real_structure, restricted, min_energy, max_energy, actual_num_str);
        f += beta*energy + logZ;
    }      
    return f;                                                                                                                                
}


void compute_gradient_f (char *input_file, double *f_gradient)
// PRE:  the parameters have been read
// POST: compute the gradient of f, which is 1/RT sum_{i=1}^N (c_{nat}^i - sum_k c_k^i exp (-1/RT G_{i,k,theta}) / sum_k exp (-1/RT G_{i,k,theta}))
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    char buffer [5000];
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double min_energy, max_energy;      
    int i, actual_num_str;
    int counter [MAXNUMPARAMS];    
    double logZ_gradient [MAXNUMPARAMS];
    
    for (i=0; i < num_params; i++)
    {
        f_gradient[i] = 0;
    }    
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }     
    int k;
    k = 0;
    while (!feof (file))
    {
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;    
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        count_each_structure_type (sequence, real_structure, counter);
        simfold_restricted_logZ_gradient (sequence, real_structure, restricted, logZ_gradient);
        for (i = 0; i < num_params; i++)
        {
            f_gradient[i] += counter[i] - logZ_gradient[i];
        }
        //printf ("counter[i]=%d, logZ_gradient[i] = %e, f_gradient[i] = %e\n", counter[7621], logZ_gradient[7621], f_gradient[7621]);
    }      
    for (i = 0; i < num_params; i++)
    {    
        f_gradient[i] *= beta;
    }    
}



double compute_f_and_gradient_f (char *input_file, double *f_gradient)
// PRE:  the parameters have been read
// POST: computes the f function, which is sum_{i=1}^N (1/RT G_{i,nat,theta} + logZ)
// POST: compute the gradient of f, which is 1/RT sum_{i=1}^N (c_{nat}^i - sum_k c_k^i exp (-1/RT G_{i,k,theta}) / sum_k exp (-1/RT G_{i,k,theta}))
{
    char sequence[MAXSLEN];
    char real_structure[MAXSLEN];
    char restricted[MAXSLEN];
    char buffer [5000];
    FILE *file;
    double logZ;
    double energy;
    double R, temp, beta;
    R = 0.00198717;
    temp = 310.15;
    beta = 1/(R*temp); 
    double min_energy, max_energy;      
    int i, actual_num_str;
    int counter [MAXNUMPARAMS];    
    double logZ_gradient;
    double f;
    f = 0;
    int real_str_found = 0;
    double numerator [MAXNUMPARAMS];
    
    double denominator;      
    char structure[MAXSLEN];
    double enthalpy;
    char tmp_structures[MAXSUBSTR][MAXSLEN];
    double tmp_energies[MAXSUBSTR];    
    
    for (i=0; i < num_params; i++)
    {
        f_gradient[i] = 0;
    }    
    if ((file = fopen (input_file, "r")) == NULL)
    {
        printf ("Cannot open file %s\n", input_file);
        exit (0);
    }     
    int k;
    k = 0;
    while (!feof (file))
    {
        if (!get_info_from_file (file, sequence, real_structure, restricted)) 
            continue;    
        //printf ("Seq: |%s|\nStr: |%s|\nRes: |%s|\n", sequence, real_structure, restricted);
        count_each_structure_type (sequence, real_structure, counter);

        // now fold sequence
        s_min_folding *min_fold = new s_min_folding (sequence, restricted);
        min_energy = min_fold->s_simfold_restricted ();
        min_fold->return_structure (structure);
        delete min_fold;      
        
        s_sub_folding* sub_fold = new s_sub_folding(sequence, restricted, -(int)(min_energy*100.0));
        sub_fold->set_limit (MAXSUBSTR);
        sub_fold->s_simfold_restricted (enthalpy);
        actual_num_str = sub_fold->return_structures(tmp_structures, tmp_energies);
        max_energy = tmp_energies[actual_num_str-1];
        delete sub_fold;
        
        for (i=0; i < num_params; i++)
        {
            numerator[i] = 0;
        }    
            
        // first compute the denominator, which is the same for all parameters
        denominator = 0;
        for (k=0; k < actual_num_str; k++)
        {
            // check if the real structure is one of the suboptimal ones
            if (strcmp (tmp_structures[k], real_structure) == 0)
                real_str_found = 1;
            // recompute the free energy, i.e. with the correct dangling ends
            energy = free_energy_simfold_restricted (sequence, tmp_structures[k], restricted);        
            denominator += exp ((-1) * energy * beta);
            count_each_structure_type (sequence, tmp_structures[k], counter);
            for (i=0; i < num_params; i++)
            {
                numerator[i] += counter[i] * exp ((-1) * energy * beta);
            }
        }
        count_each_structure_type (sequence, real_structure, counter);
        energy = free_energy_simfold_restricted (sequence, real_structure, restricted);        
        if (!real_str_found)
        {
            // if not found, add it
            denominator += exp ((-1) * energy * beta);                    
            for (i=0; i < num_params; i++)
            {
                numerator[i] += counter[i] * exp ((-1) * energy * beta);
            }    
        }
        // now the denominator and nominator are computed
        
        f += beta*energy + log(denominator);
        //printf ("logZ = %e\n", log(denominator));
        for (i=0; i < num_params; i++)
        {
            logZ_gradient = numerator[i] / denominator;
            f_gradient[i] += counter[i] - logZ_gradient;
        }
        //printf ("counter[i]=%d, logZ_gradient[i] = %e, f_gradient[i] = %e\n", counter[7621], numerator[7621]/denominator, f_gradient[7621]);
        //printf ("logZ_grad[i] = %e\n", numerator[7621]/denominator);
    }      
    for (i = 0; i < num_params; i++)
    {    
        f_gradient[i] *= beta;
    }    
    return f;
}
